#include <stdio.h>

#ifndef ATARI_ST

/*
 * IBM version
 */
#include <conio.h>
#include <bios.h>

#else

/*
 * Atari ST version - Mark Williams C
 *   port by Pat McElhatton   April 27 1992
 *
 * build with: cc samdisk.c -DATARI_ST -o samdisk.ttp
 */
#include <ctype.h>
#include <osbind.h>
#include <bios.h>
#include <errno.h>

#endif

#ifndef TRUE
#define TRUE 1
#define FALSE !TRUE
#endif

#define NSECS 9
#define NTRACKS 80
#define NBYTES 512
#define NHEADS 2
#define SEEK0 17 	 /* RESEEK to track 0 */
#define SECREAD 2	/* BIOSDISK read # of sectors command */
#define SECWRITE 3
#define STARTSEC 1
#define DEFDRIVE 01	/* assume drive B: for now */
#ifndef ATARI_ST
#define DISKCHANGE 6	/* BIOSDISK() disk change indicator */
#define DISKNREADY 128
#else
#define DISKCHANGE AE_CHNG	/* BIOSDISK() disk change indicator */
#define DISKNREADY AEDRVNR
#endif
#define READMODE 1
#define WRITEMODE 2
#define LISTMODE 3

#define S330 "S330"
#define S50 "S-50"
#define S550 "S550"

#define MODES330 1
#define MODES50 2
#define MODES550 3

#define	TRACKSIZE NSECS
#define	CYLSIZE TRACKSIZE*2

#define PATCHBASE (0xFC00/NBYTES)   /* S330 sector with start of patchnames */
#define PATCHOFFSET 0x100
#define PATCHQTY 16
#define PNAMESIZE 12

#define TONEBASE (0x11E00/NBYTES)  /* S330 sector with TONE names */
#define TONEOFFSET 0x10
#define TONEQTY 32
#define TNAMESIZE 8

#define TITLE_SECTOR 0x86	/* Sector with titles in it */
#define TITLE_OFFSET 0x33
#define TITLE_LINE_SZ 12

#define DEFFILE "samdisk.out"

int headndx, trackndx, result, catalogue = FALSE, replicate=FALSE;
int drivenum = DEFDRIVE, dmode = 0, binmode = FALSE;
unsigned char secbuff[NBYTES];

char patches[PATCHQTY][PNAMESIZE+1];
char tones[TONEQTY][TNAMESIZE+1];
char title[TITLE_LINE_SZ+1];
char stitle[(TITLE_LINE_SZ+1)*4];
char modename[5];
char *filename = DEFFILE;

unsigned char trackbuffer[NBYTES*NSECS];

FILE *ifp, *ofp;

#ifndef ATARI_ST
extern int errno;
extern int optind, opterr;
extern char *optarg;

#else

#define CONPORT 2
/*
 * consoleOut - write a character to console
 */
void consoleOut( ch )
    char ch;
{
    Bconout( CONPORT, ch );
}


#define ESCAPE 0x1B

clreol()
{
    /* erase to end of line */
    consoleOut( ESCAPE );
    consoleOut( 'K' );
}

#define cprintf printf

gotoxy( x, y )
    int x, y;
{
    /* position the cursor */
    consoleOut( ESCAPE );
    consoleOut( 'Y' );
    consoleOut( y + 31 );
    consoleOut( x + 31 );	
}

char getch()
{
    return ( Bconin(CONPORT) & 0x7f );
}

clrscr()
{
    /* move cursor to top line */
    gotoxy( 1, 1 );

    /* erase to end of screen */
    consoleOut( ESCAPE );
    consoleOut( 'J' );
}

#endif

main(argc,argv)
int argc;
char *argv[];
	{

	int mode = READMODE;
	int option, i;

#ifndef ATARI_ST
	while((option = getopt(argc,argv,"bclrd:f:w?")) != EOF)
		switch(tolower(option))
			{
			case 'd':
				drivenum = toupper(*(optarg))-'A';
				if(drivenum > 3)
					drivenum = 3; /* max four floppies */
				break;

			case 'f':
				filename = optarg;
				break;

			case 'r':
				replicate = TRUE;
				break;

			case 'l':
				mode = LISTMODE;
				break;

			case 'c':
				catalogue = TRUE;
				mode = LISTMODE;
				break;

			case 'b':
				binmode = TRUE;
				catalogue = TRUE;
				mode = LISTMODE;
				break;

			case 'w':
				mode = WRITEMODE;
				break;

			case '?':
			default :
				usage();

			}

	if((argc > 1) && (optind > argc))
		usage();
#else
	int holdmode;
	holdmode = FALSE;

	/* turn off buffering on standard out */
	setbuf( stdout, NULL );

	for ( i = 1; i < argc; i++ ) {
		option = *(argv[i] + 1);
		switch(tolower(option))
			{
			case 'd':
				drivenum = toupper(*(argv[i] + 2))-'A';
				if(drivenum > 3)
					drivenum = 3; /* max four floppies */
				break;

			case 'f':
				filename = argv[i] + 2;
				break;

			case 'r':
				replicate = TRUE;
				break;

			case 'l':
				mode = LISTMODE;
				break;

			case 'c':
				catalogue = TRUE;
				mode = LISTMODE;
				break;

			case 'b':
				binmode = TRUE;
				catalogue = TRUE;
				mode = LISTMODE;
				break;

			case 'w':
				mode = WRITEMODE;
				break;

			case 'h':
				holdmode = TRUE;
				break;

			case '?':
			default :
				usage();

			}
	}
#endif

	if(replicate)
		{
		dupe_mode();
		exit(0);
		}

	if(mode == LISTMODE)
	{

	/* Let's get the PATCH names */

	ofp = stdout;

	if(catalogue)
	   if(strcmp(filename, DEFFILE) != 0)
			if((ofp = fopen(filename, binmode?"ab":"a")) == (FILE *)NULL)
				{
				fprintf(stderr,"Cannot open %s file for append - %s\n",
				binmode ? "database" : "catalogue", filename);
				perror("");
				exit(errno);
				}

	if((binmode) && (ofp == stdout))
		{
		fprintf(stderr,"You _must_ specify a filename for binary record mode\n");
		exit(1);
		}

	do
	{
	dmode = id_disk();

	if(dmode >= 0)
	{
	getpnames();
	gettnames();
	get_title();
	if(binmode)
		dump_data();
	else
		print_report();
	}

	if(catalogue)
		catalogue = get_prompt();

	} while (catalogue);

	if((catalogue)  && (ofp != stdout))
		fclose(ofp);


	}
	else
	{

	if(mode == READMODE)
		read_disk();
	else
		write_disk();



	     }
#ifdef ATARI_ST
	  if ( holdmode == TRUE ) {
		printf ( "hit any key to exit" );
		i = getch();
	  }
	  exit(0);
#endif
	}





usage()
	{
#ifdef ATARI_ST
	fprintf(stderr,"\nUsage: samdisk [-wlrcbh] [-dFLOPPY] [-fFILENAME] \n");
#else
	fprintf(stderr,"\nUsage: samdisk [-wlrcb] [-dFLOPPY] [-fFILENAME] \n");
#endif
	fprintf(stderr,"\nDEFAULT    Read mode; sampler disks are read; output to diskfile '%s'.\n\n", DEFFILE);
	fprintf(stderr,"-w         Write mode. Writes from a disk file to a sampler floppy.\n");
	fprintf(stderr,"-l         List mode. Prints patch and tone contents.\n");
	fprintf(stderr,"-r         Replicate mode. Makes one or more copies of a source diskette.\n");
	fprintf(stderr,"-c         Catalogue mode. Repeatedly asks for more disks; Diskfile or screen.\n");
	fprintf(stderr,"-b         Binary catalogue. Writes binary records for manipulation by other\n");
	fprintf(stderr,"           programs of your own design.\n");
	fprintf(stderr,"-dFLOPPY   Specify another floppy disk (default is B:).\n");
	fprintf(stderr,"-fFILENAME Specify another filename (default is '%s').\n", DEFFILE);
#ifdef ATARI_ST
	fprintf(stderr,"-h         Hold mode. Waits for keypress before exiting.\n" );
#endif
	exit(1);
	}


getpnames()
	{
	int count, count2;
	char *wptr;

	/* init patchnames */

	for(count = 0; count < PATCHQTY; count++)
		patches[count][0] = NULL;

	for(count = PATCHBASE; count < (PATCHBASE+(PATCHQTY/2)); count++)
		{
		tryagain:
		result = getsect(count, secbuff);
		if(result)
			{
			if(result != DISKCHANGE)
				{
				fprintf(stderr,"Problem reading patchnames - biosdisk error: %d\n", result);
				exit(1);
				}
			else
				goto tryagain;
			}


		for(count2 = 0; count2 < 2; count2++)
			{
			wptr = secbuff+(count2*PATCHOFFSET);
			strncpy(patches[(count-PATCHBASE)*((dmode == MODES50) ? 1 : 2)+count2], wptr, PNAMESIZE);
			patches[(count-PATCHBASE)*2+count2][PNAMESIZE] = '\0';
			if(dmode == MODES50)
				count2++;
			}
		}


	}

gettnames()
	{
	int count, count2;
	char *wptr;

		for(count = 0; count < TONEQTY; count++)
			tones[count][0] = NULL;

		count = TONEBASE;
		result = getsect(count, secbuff);
		if(result)
			if(result != DISKCHANGE)
				{
				fprintf(stderr,"Problem reading tone names - biosdisk error: %d\n", result);
				exit(1);
				}


		for(count2 = 0; count2 < TONEQTY; count2++)
			{
			wptr = secbuff+(count2*TONEOFFSET);
			strncpy(tones[count2], wptr, TNAMESIZE);
			tones[count2][TNAMESIZE] = '\0';
			}





	}

getsect(secnum,buff)
int secnum;
unsigned char *buff;
	{
	int sector, gs_result;

	trackndx = (secnum/(CYLSIZE));
	headndx =  (secnum/NSECS)%2;
	sector =  secnum - ((trackndx*(CYLSIZE)) + (headndx * NSECS)) + 1;

	reget_abs:
#ifdef ATARI_ST
	gs_result = Floprd( buff, 0L, drivenum, sector, trackndx, headndx, 1 );
#else
	gs_result = (biosdisk(SECREAD, drivenum, headndx, trackndx, sector, 1, buff));
#endif
	if(gs_result == DISKNREADY)
		{
		if(ckerr(gs_result))
			goto reget_abs;
		}
	else
		return(gs_result);
	}

fixnum(num)
int num;
	{
	int tens = num/8+1;
	int ones = num%8 +1;
	return(tens*10 + ones);
	}

get_title()
	{
	char  *wptr;
	int row, column;

	*title = NULL;
	*stitle = NULL;

	result = getsect(TITLE_SECTOR,secbuff);

	strncpy(title,secbuff+TITLE_OFFSET, TITLE_LINE_SZ);
	title[TITLE_LINE_SZ] = '\0';

	/* get subtitle */

	wptr = secbuff+TITLE_OFFSET+TITLE_LINE_SZ;


	for(column = 0; column < TITLE_LINE_SZ; column++)
		for(row = 0; row < 4; row++)
			stitle[row*(TITLE_LINE_SZ+1)+column] = *wptr++;
        	for(row = 0; row < 4; row++)
			stitle[row*(TITLE_LINE_SZ+1)+TITLE_LINE_SZ] = ' ';
		stitle[4*(TITLE_LINE_SZ+1)-1] = NULL;



	}

print_report()
	{
	int count;

	fprintf(ofp,"\n===============================================================================\nTitle: %-12s - %-48s\n Type: %-4s\n", title, stitle, modename);
	fprintf(ofp,"-------------------------------------------------------------------------------\n");

	for(count = 0; count < PATCHQTY; count++)
		fprintf(ofp,"Patch P%02d: %-12s      Tone T%02d: %-8s      Tone T%02d: %-8s\n",
		fixnum(count),patches[count], fixnum(count), tones[count],
		fixnum(count+16), tones[count+16]);


	}
get_prompt()
	{
	char c;

	if(ofp != stdout)
		{
		clrscr();
		fprintf(stderr,"\n[%-4s] %-12s %-52s - DONE\n", modename, title, stitle);
		}



	fprintf(stderr,"\nPlease insert next disk and hit any key, or hit ESC to quit: \n");
	c = getch();
	if(ofp != stdout)
		clrscr();

	if(c == 0x1b)
		return(FALSE);
	else
		return(TRUE);

	}

id_disk()
	{
	char *wptr = &secbuff[4];

	another_try:
	result = getsect(0,secbuff);


	if(result != 0)
		if(result == DISKCHANGE)
			goto another_try;
		else
			{
			fprintf(stderr,"Problem ID'ing disk - bioserr: %d\n", result);
			return(-1);
			}

	secbuff[8] = 0;

	if(strcmp(wptr,S330) == 0)
		{
		strcpy(modename,S330);
		return(MODES330);
		}
	if(strcmp(wptr,S50) == 0)
		{
		strcpy(modename,S50);
		return(MODES50);
		}
	if(strcmp(wptr,S550) == 0)
		{
		strcpy(modename,S550);
		return(MODES550);
		}
	strcpy(modename, "????");
	return(0);


	}

dump_data()
	{
	int error = 0;

	error |= (fwrite(modename, sizeof modename, 1 , ofp) != 1);
	error |= (fwrite(title, sizeof title, 1, ofp) != 1);
	error |= (fwrite(stitle, sizeof stitle, 1, ofp) != 1);
	error |= (fwrite(patches, sizeof patches, 1, ofp) != 1);
	error |= (fwrite(tones, sizeof tones, 1, ofp) != 1);

	if(error)
		{
		fprintf(stderr,"Problems writing to database file - ");
		perror("");
		exit(errno);
		}
	}

read_disk()
	{

		/* Create HOLDING file */

		if((ofp = fopen(filename, "wb")) == (FILE *)NULL)
			{
			fprintf(stderr,"Cannot open output file - %s\n", filename);
			perror("");
			exit(errno);
			}

		clrscr();
		dmode = id_disk();
		if(dmode >= 0)
		{
		get_title();
		gotoxy(1,10);
		cprintf("Title: %-12s - %-48s\n\rType : %-4s\n\rFile : %s", title, stitle, modename, filename);

		for(trackndx = 0; trackndx < NTRACKS; trackndx++)
			for(headndx = 0; headndx < NHEADS; headndx++)
			{
			gotoxy(1,14);
			cprintf("Reading Track %02d  Head %d",trackndx, headndx);
		reread:
#ifdef ATARI_ST
			result = Floprd( trackbuffer, 0L, drivenum, 1, trackndx, headndx, NSECS );
#else
			result = biosdisk(SECREAD, drivenum, headndx, trackndx, 1, NSECS, trackbuffer);
#endif

			if(result)
				if(ckerr(result))
					goto reread;

			if(fwrite(trackbuffer,NBYTES*NSECS,1,ofp) != 1)
				fprintf(stderr,"Fwrite problem\n");
			}
		}
		fclose(ofp);
		gotoxy(1,16);
	}

write_disk()
	{

		dmode = id_disk();	/* a dummy for synch */
		if(dmode >=0)
		{
		/* open SOURCE file */

		if((ifp = fopen(filename, "rb")) == (FILE *)NULL)
			{
			fprintf(stderr,"Cannot open input file - %s\n", filename);
			perror("");
			exit(errno);
			}

                clrscr();
		gotoxy(1,12);
		cprintf("File : %s", filename);


                for(trackndx = 0; trackndx < NTRACKS; trackndx++)
			for(headndx = 0; headndx < NHEADS; headndx++)
			{
			if(fread(trackbuffer,NBYTES*NSECS,1,ifp) != 1)
				{
				fprintf(stderr,"\nFread problem - possibly improper file size.\n");
				exit(errno);
				}

			gotoxy(1,14);
			cprintf("Writing Track %02d  Head %d",trackndx, headndx);
		  rewrite:
#ifdef ATARI_ST
			result = Flopwr( trackbuffer, 0L, drivenum, 1, trackndx, headndx, NSECS );
#else
			result = biosdisk(SECWRITE, drivenum, headndx, trackndx, 1, NSECS, trackbuffer);
#endif

			if(result)
				if(ckerr(result))
					goto rewrite;


			}
		fclose(ifp);

                dmode = id_disk();
		get_title();
		gotoxy(1,10);
		cprintf("Title: %-12s - %-48s\n\rType : %-4s", title, stitle, modename);
		gotoxy(1,16);
		}
	}

ckerr(code)
	{
	char response;

	switch(code)
		{
		case DISKCHANGE:
				return(6);

		case DISKNREADY:
				gotoxy(1,25);
				clreol();
				cprintf("Disk not ready! Press any key to retry, or ESC to abort: ");
				response = getch();
				gotoxy(1,25);
				clreol();
				if(response == 0x1b)
					exit(1);
				else
					return(1);
		default:
			fprintf(stderr,"Bios disk error: %d - aborting\n", code);
			exit(1);
		}

	}

dupe_mode()
	{
	int do_another = FALSE;

	read_disk();

	ask_for_more(0);

		do
		{
		write_disk();
		do_another = ask_for_more(1);
		} while (do_another);


	}


ask_for_more(flag)
int flag;
	{
	char response;

	if(flag == 1)
	{
	gotoxy(1,25);
	clreol();
	cprintf("Make another copy of this disk?");
	response = tolower(getch());
	}

	gotoxy(1,25);
	clreol();

	if((response == 'y') || (flag == 0))
		{
		cprintf("Insert %s blank 720K disk and press any key, or ESC to abort: ", flag ? "next" : "a");
		response = tolower(getch());
		gotoxy(1,25);
		clreol();

		if(response != 0x1b)
			return(TRUE);
		else
			return(FALSE);
		}

	else
		return(FALSE);
	}
